import dominix
from dominix.tags import *
from dominix import util


def test_context():
  id1 = dominix.dom_tag._get_thread_context()
  id2 = dominix.dom_tag._get_thread_context()
  assert id1 == id2

def test_include():
  import os
  try:
    f = open('_test_include.deleteme', 'w')
    f.write('Hello World')
    f.close()

    d = div()
    d += util.include('_test_include.deleteme')
    assert d.render() == '<div>Hello World</div>'

  finally:
    try:
      os.remove('_test_include.deleteme')
    except:
      pass


def test_system():
  d = div()
  d += util.system('echo Hello World')
  assert d.render().replace('\r\n', '\n') == '<div>Hello World\n</div>'


def test_unescape():
  assert util.unescape('&amp;&lt;&gt;&#32;') == '&<> '


def test_url():
  assert util.url_escape('hi there?') == 'hi%20there%3F'
  assert util.url_unescape('hi%20there%3f') == 'hi there?'


def test_container():
  d = div()
  with d:
    with util.container():
      pass
  assert d.render() == '<div></div>'


  d = div()
  with d:
    with util.container():
      h1('a')
  assert d.render() == \
'''<div>
  <h1>a</h1>
</div>'''



import pytest

from dominix.tags import *

def test_dom():
  container = div()
  with container.add(div(id='base')) as dom:
    s1 = span('Hello', id='span1')
    s2 = span('World', id='span2')

  s3 = span('foobar', id='span3')
  dom.appendChild(s3)
  assert container.getElementById('base') is dom
  assert container.getElementById('span1') is s1
  assert container.getElementById('span3') is s3
  assert container.getElementById('foo') is None
  assert container.getElementsByTagName('span') == [s1, s2, s3]
  assert container.getElementsByTagName('SPAN') == [s1, s2, s3]
  assert container.getElementsByTagName(1234) is None

def test_element():
  d = div(
    span(id='a'),
    span(id='a'),
  )
  with pytest.raises(ValueError):
    d.getElementById('a')

def test_parent_node():
  parent = div(id='parent')
  child = div(id='child')
  parent.add(child)

  assert child.parentNode is parent



__license__ = '''
This file is part of Dominix.

Dominix is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

Dominix is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with Dominix.  If not, see
<http://www.gnu.org/licenses/>.
'''
# pylint: disable=bad-whitespace

from setuptools import setup

import imp
_version = imp.load_source("dominix._version", "dominix/_version.py")

long_description = open('README.md').read()

setup(
  name    = 'dominix',
  version = _version.__version__,
  author  = 'Tom Flanagan and Jake Wharton',
  author_email = 'tom@zkpq.ca',
  license = 'LGPLv3',
  url     = 'https://github.com/Knio/dominix/',
  description      = 'Dominix is a Python library for creating and manipulating HTML documents using an elegant DOM API.',
  long_description = long_description,
  long_description_content_type='text/markdown',
  keywords         = 'framework templating template html xhtml python html5',

  python_requires='>=2.7, <3',
  classifiers = [
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: Implementation :: PyPy',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Text Processing :: Markup :: HTML',
  ],

  packages = ['dominix'],
  include_package_data = True,
)



import pytest
try:
    import mock # type: ignore
except ImportError:
    import unittest.mock as mock

from dominix.tags import *


def test___get_thread_context(monkeypatch):
    from dominix import dom_tag as sut
    greenlet = mock.Mock()
    greenlet.getcurrent.return_value = 100
    monkeypatch.setattr(sut, 'greenlet', greenlet)

    threading = mock.Mock()
    threading.current_thread.return_value = 200
    monkeypatch.setattr(sut, 'threading', threading)

    assert sut._get_thread_context() is not None

def test_add_raw_string():
    container = div()
    container.add_raw_string('foo')
    assert container.children == ['foo']

def test_clear():
    container = div()
    child = div()
    container.add(child)

    assert container.children == [child]
    assert child.parent == container

    container.clear()

    assert container.children == []
    assert child.parent is None

def test_set_attribute():
    container = div()
    container.add_raw_string('foo')
    container.set_attribute(0, 'bar')

    assert container.children == ['bar']

def test_set_attribute_error():
    container = div()
    with pytest.raises(TypeError, match=(
            'Only integer and string types are valid for assigning '
            'child tags and attributes, respectively.'
        )):
        container.set_attribute(1.0, 'foo')

def test___get_item___child_index_error():
    d = div()
    with pytest.raises(IndexError, match='Child with index "10" does not exist.'):
        d[10]

def test___contains__():
    container = div()
    container.add(div())
    assert 'div' in container

def test_nested_context():
    def sub(*args):
        with div('B') as B:
            B.add(*args)

    with div('A') as A:
        sub(div('C'))

    assert str(A) == \
'''<div>A
  <div>B
    <div>C</div>
  </div>
</div>'''



from dominix.tags import *

def test_cls():
    tag = html()
    tag.add_class("foo")
    assert tag.render() == '<html class="foo"></html>'
    tag = html(cls="foo  bar")
    assert tag.render() == '<html class="foo  bar"></html>'
    classes = tag.cls
    tag.cls.remove("foo")
    assert tag.render() == '<html class="bar"></html>'
    classes.remove("bar")
    assert tag.render() == '<html></html>'
    classes.extend(["baz", "qux"])
    assert tag.render() == '<html class="baz qux"></html>'
    tag.cls = "kaz quux "
    assert tag.render() == '<html class="kaz quux "></html>'
    tag.add_class("foo", "bar")
    assert tag.render() == '<html class="kaz quux foo bar"></html>'
    tag.rem_class("quux", "foo", "fux")
    assert tag.render() == '<html class="kaz bar"></html>'
    tag.add_class("foo", "bar")
    tag.rem_class("bar")
    assert tag.render() == '<html class="kaz foo"></html>'
    with tag:
        add_class("bar")
        assert tag.render() == '<html class="kaz foo bar"></html>'
        rem_class("foo")
        assert tag.render() == '<html class="kaz bar"></html>'



def test_style():
    tag = html(style="foo: bar;  baz: qux")
    assert tag.render() == '<html style="foo: bar;  baz: qux"></html>'
    style = tag.style
    del style["foo"]
    assert tag.render() == '<html style="baz:qux"></html>'
    tag.style.update({"foo": "bar", "baz": "zab"})
    assert tag.render() == '<html style="baz:zab; foo:bar"></html>'
    tag.upd_style("foo", "qux").upd_style({"baz": "bar", "qux": "foo"})
    assert tag.render() == '<html style="baz:bar; foo:qux; qux:foo"></html>'
    tag.del_style("foo", "wtf").del_style("qux")
    assert tag.render() == '<html style="baz:bar"></html>'
    with tag:
        upd_style("foo", "bar")
        assert tag.render() == '<html style="baz:bar; foo:bar"></html>'
        del_style("foo")
        assert tag.render() == '<html style="baz:bar"></html>'

def test_hx_on():
    tag = html(hx_on_click="foo()")
    assert tag.render() == '<html hx-on-click="foo()"></html>'
    tag = html(hx_on={"click": "foo()"})
    assert tag.hx_on["click"] == "foo()"
    tag = html()
    tag.hx_on["click"] = "foo()"
    assert tag.render() == '<html hx-on:click="foo()"></html>'
    tag.hx_on["click"] = "bar()"
    assert tag.render() == '<html hx-on:click="bar()"></html>'
    tag.hx_on.update({"htmx:before-request": "baz()", "htmx:after-request": "qux()"})
    assert tag.render() == '<html hx-on:click="bar()" hx-on:htmx:after-request="qux()" hx-on:htmx:before-request="baz()"></html>'
    assert tag.hx_on == {"click": "bar()", "htmx:before-request": "baz()", "htmx:after-request": "qux()"}
    del tag.hx_on["click"]
    assert tag.render() == '<html hx-on:htmx:after-request="qux()" hx-on:htmx:before-request="baz()"></html>'
    tag.hx_on.update({"htmx:before-request": "foo()", "htmx:after-request": None})
    assert tag.render() == '<html hx-on:htmx:before-request="foo()"></html>'
    tag.hx_on.clear()
    assert tag.render() == '<html></html>'
    with tag:
        upd_hx_on("click", "foo()")
        assert tag.render() == '<html hx-on:click="foo()"></html>'
        del_hx_on("click")
        assert tag.render() == '<html></html>'

def test_hx_vals():
    tag = html(hx_vals='{ "foo": "bar", "baz": "qux" }')
    del tag.hx_vals["foo"]
    assert tag.render() == '''<html hx-vals='{"baz": "qux"}'></html>'''
    tag = html()
    tag.hx_vals.update({"foo": "bar", "baz": "qux"})
    assert tag.hx_vals == {"foo": "bar", "baz": "qux"}
    assert tag.hx_vals["foo"] == "bar"
    assert tag.hx_vals.get("bazz", None) == None
    assert tag.render() == '''<html hx-vals='{"foo": "bar", "baz": "qux"}'></html>'''
    del tag.hx_vals["foo"]
    assert tag.render() == '''<html hx-vals='{"baz": "qux"}'></html>'''
    tag.hx_vals.clear()
    assert tag.render() == '<html></html>'
    tag.upd_hx_vals("foo", "bar").upd_hx_vals({"baz": "qux", "qux": "foo"})
    assert tag.render() == '<html hx-vals=\'{"foo": "bar", "baz": "qux", "qux": "foo"}\'></html>'
    tag.del_hx_vals("foo", "wtf").del_hx_vals("qux")
    assert tag.render() == '<html hx-vals=\'{"baz": "qux"}\'></html>'
    with tag:
        upd_hx_vals("foo", "bar")
        assert tag.render() == '<html hx-vals=\'{"baz": "qux", "foo": "bar"}\'></html>'
        del_hx_vals("foo")
        assert tag.render() == '<html hx-vals=\'{"baz": "qux"}\'></html>'

def test_hx_headers():
    tag = html(hx_headers='''{"foo": "bar", "baz": "qux"}''')
    assert tag.hx_headers == {"foo": "bar", "baz": "qux"}
    tag = html(hx_headers={"foo": "bar", "baz": "qux"})
    assert tag.hx_headers == {"foo": "bar", "baz": "qux"}
    assert tag.hx_headers["foo"] == "bar"
    assert tag.hx_headers.get("bazz", None) == None
    assert tag.render() == '''<html hx-headers='{"foo": "bar", "baz": "qux"}'></html>'''
    del tag.hx_headers["foo"]
    assert tag.render() == '''<html hx-headers='{"baz": "qux"}'></html>'''
    tag.hx_headers.clear()
    assert tag.render() == '<html></html>'
    tag.upd_hx_headers("foo", "bar").upd_hx_headers({"baz": "qux", "qux": "foo"})
    assert tag.render() == '<html hx-headers=\'{"foo": "bar", "baz": "qux", "qux": "foo"}\'></html>'
    tag.del_hx_headers("foo", "wtf").del_hx_headers("qux")
    assert tag.render() == '<html hx-headers=\'{"baz": "qux"}\'></html>'
    with tag:
        upd_hx_headers("foo", "bar")
        assert tag.render() == '<html hx-headers=\'{"baz": "qux", "foo": "bar"}\'></html>'
        del_hx_headers("foo")
        assert tag.render() == '<html hx-headers=\'{"baz": "qux"}\'></html>'

def test_hx_get():
    tag = html()
    tag.hx_get = "bar"
    assert tag.render() == '<html hx-get="bar"></html>'

def test_all_args():
    tag = html(
        hx_get="a",
        hx_post="b",
        hx_push_url="d",
        hx_select="e",
        hx_select_oob="f",
        hx_swap="g",
        hx_swap_oob="h",
        hx_target="i",
        hx_trigger="j",
        hx_vals={"k": "l"},
        hx_boost=True,
        hx_confirm="m",
        hx_delete="n",
        hx_disable="o",
        hx_disable_elt="p",
        hx_disinherit="q",
        hx_encoding="r",
        hx_ext="s",
        hx_headers={"t": "u"},
        hx_history=False,
        hx_history_elt="v",
        hx_include="w",
        hx_indicator="x",
        hx_params="z",
        hx_patch="aa",
        hx_preserve="ab",
        hx_prompt="ac",
        hx_put="ad",
        hx_replace_url="ae",
        hx_request="af",
        hx_sync="ag",
        hx_validate="ah",
    )
    assert tag.render() == '''<html hx-boost="true" hx-confirm="m" hx-delete="n" hx-disable="o" hx-disable-elt="p" hx-disinherit="q" hx-encoding="r" hx-ext="s" hx-get="a" hx-headers='{"t": "u"}' hx-history="false" hx-history-elt="v" hx-include="w" hx-indicator="x" hx-params="z" hx-patch="aa" hx-post="b" hx-preserve="ab" hx-prompt="ac" hx-push-url="d" hx-put="ad" hx-replace-url="ae" hx-request="af" hx-select="e" hx-select-oob="f" hx-swap="g" hx-swap-oob="h" hx-sync="ag" hx-target="i" hx-trigger="j" hx-validate="ah" hx-vals='{"k": "l"}'></html>'''

def test_all_props():
    tag = html()
    tag.hx_get="a"
    tag.hx_post="b"
    tag.hx_push_url="d"
    tag.hx_select="e"
    tag.hx_select_oob="f"
    tag.hx_swap="g"
    tag.hx_swap_oob="h"
    tag.hx_target="i"
    tag.hx_trigger="j"
    tag.hx_vals={"k": "l"}
    tag.hx_boost=True
    tag.hx_confirm="m"
    tag.hx_delete="n"
    tag.hx_disable="o"
    tag.hx_disable_elt="p"
    tag.hx_disinherit="q"
    tag.hx_encoding="r"
    tag.hx_ext="s"
    tag.hx_headers={"t": "u"}
    tag.hx_history=False
    tag.hx_history_elt="v"
    tag.hx_include="w"
    tag.hx_indicator="x"
    tag.hx_params="z"
    tag.hx_patch="aa"
    tag.hx_preserve="ab"
    tag.hx_prompt="ac"
    tag.hx_put="ad"
    tag.hx_replace_url="ae"
    tag.hx_request="af"
    tag.hx_sync="ag"
    tag.hx_validate="ah"
    assert tag.render() == '''<html hx-boost="true" hx-confirm="m" hx-delete="n" hx-disable="o" hx-disable-elt="p" hx-disinherit="q" hx-encoding="r" hx-ext="s" hx-get="a" hx-headers='{"t": "u"}' hx-history="false" hx-history-elt="v" hx-include="w" hx-indicator="x" hx-params="z" hx-patch="aa" hx-post="b" hx-preserve="ab" hx-prompt="ac" hx-push-url="d" hx-put="ad" hx-replace-url="ae" hx-request="af" hx-select="e" hx-select-oob="f" hx-swap="g" hx-swap-oob="h" hx-sync="ag" hx-target="i" hx-trigger="j" hx-validate="ah" hx-vals='{"k": "l"}'></html>'''


def test_attr():
    tag = html()
    tag.attr(href="foo")
    assert tag.render() == '<html href="foo"></html>'
    with tag as t:
        attr(cls="bar")
        t.attr(href="bax")
    assert tag.render() == '<html class="bar" href="bax"></html>'
    tag.attr(cls="baz", href="qux", id="quux")
    assert tag.render() == '<html class="baz" href="qux" id="quux"></html>'

def test_quotes():
    tag = div(dbl='b"c', sng="a'b", both="""a"'b""")
    assert tag.render() == '''<div both="a&quot;'b" dbl='b"c' sng="a'b"></div>'''



from asyncio import gather, run, Semaphore
from dominix.dom_tag import async_context_id
from textwrap import dedent

from dominix import tags

# To simulate sleep without making the tests take a hella long time to complete
# lets use a pair of semaphores to explicitly control when our coroutines run.
# The order of execution will be marked as comments below:
def test_async_bleed():
    async def tag_routine_1(sem_1, sem_2):
        root = tags.div(id = 1) # [1]
        with root: # [2]
            sem_2.release() # [3]
            await sem_1.acquire() # [4]
            tags.div(id = 2) # [11]
        return str(root) # [12]

    async def tag_routine_2(sem_1, sem_2):
        await sem_2.acquire() # [5]
        root = tags.div(id = 3) # [6]
        with root: # [7]
            tags.div(id = 4) # [8]
        sem_1.release() # [9]
        return str(root) # [10]

    async def merge():
        sem_1 = Semaphore(0)
        sem_2 = Semaphore(0)
        return await gather(
            tag_routine_1(sem_1, sem_2), 
            tag_routine_2(sem_1, sem_2)
        )

    # Set this test up for failure - pre-set the context to a non-None value.
    # As it is already set, _get_async_context_id will not set it to a new, unique value
    # and thus we won't be able to differentiate between the two contexts. This essentially simulates
    # the behavior before our async fix was implemented (the bleed):
    async_context_id.set(0)
    tag_1, tag_2 = run(merge())

    # This looks wrong - but its what we would expect if we don't
    # properly handle async...
    assert tag_1 == dedent("""\
        <div id="1">
          <div id="3">
            <div id="4"></div>
          </div>
          <div id="2"></div>
        </div>
    """).strip()

    assert tag_2 == dedent("""\
        <div id="3">
          <div id="4"></div>
        </div>
    """).strip()

    # Okay, now lets do it right - lets clear the context. Now when each async function
    # calls _get_async_context_id, it will get a unique value and we can differentiate.
    async_context_id.set(None)
    tag_1, tag_2 = run(merge())

    # Ah, much better...
    assert tag_1 == dedent("""\
        <div id="1">
          <div id="2"></div>
        </div>
    """).strip()
    
    assert tag_2 == dedent("""\
        <div id="3">
          <div id="4"></div>
        </div>
    """).strip()



from dominix import document
from dominix.tags import *

def test_doc():
  d = document()
  assert d.render() == \
'''<!DOCTYPE html>
<html>
  <head>
    <title>Dominix</title>
  </head>
  <body></body>
</html>'''


def test_decorator():
  @document()
  def foo():
    p('Hello World')

  f = foo()
  assert f.render() == \
'''<!DOCTYPE html>
<html>
  <head>
    <title>Dominix</title>
  </head>
  <body>
    <p>Hello World</p>
  </body>
</html>'''


def test_bare_decorator():
  @document
  def foo():
    p('Hello World')

  assert foo().render() == \
'''<!DOCTYPE html>
<html>
  <head>
    <title>Dominix</title>
  </head>
  <body>
    <p>Hello World</p>
  </body>
</html>'''


def test_title():
  d = document()
  assert d.title == 'Dominix'

  d = document(title='foobar')
  assert d.title == 'foobar'

  d.title = 'baz'
  assert d.title == 'baz'

  d.title = title('bar')
  assert d.title == 'bar'

  assert d.render() == \
'''<!DOCTYPE html>
<html>
  <head>
    <title>bar</title>
  </head>
  <body></body>
</html>'''

def test_containers():
  d = document()
  with d.footer:
    div('footer')
  with d:
    div('main1')
  with d.main:
    div('main2')
  print(d.header)
  print(d)
  print(d.body.children)
  with d.header:
    div('header1')
    div('header2')
  assert d.render() == \
'''<!DOCTYPE html>
<html>
  <head>
    <title>Dominix</title>
  </head>
  <body>
    <div>header1</div>
    <div>header2</div>
  ''''''
    <div>main1</div>
    <div>main2</div>
  ''''''
    <div>footer</div>
  </body>
</html>'''


def test_attributes():
  d = document(title=None, doctype=None, lang='en')
  assert d.render() == \
'''<html lang="en">
  <head></head>
  <body></body>
</html>'''

def test_repr():
  d = document(title='foo')
  assert d.__repr__() == '<dominix.document "foo">'

if __name__ == '__main__':
  # test_doc()
  test_decorator()
